from cli.utils import download_url, fetch_auth, parse_filter
from zpy.files import to_pathlib_path
import json
import requests

DATASET_TYPES = ['uploaded-data-sets', 'generated-data-sets', 'job-data-sets']


@fetch_auth
def filter_datasets(dfilter, url, auth_headers):
    """ filter datasets

    Filter dataset objects on ZumoLabs backend by given dfilter.
    Parse dfilter using parse_filter and make calls to 
    generated, uploaded, job dataset endpoints.

    Args:
        dfilter (str): filter query for datasets
        url (str): backend endpoint
        auth_headers: authentication for backend

    Return:
        dict: filtered datasets by dfilter {'name': 'id'}
    """
    filtered_datasets = {}
    field, pattern, regex = parse_filter(dfilter)
    for dataset_type in DATASET_TYPES:
        endpoint = f'{url}/api/v1/{dataset_type}/?{field}__{pattern}={regex}'
        while endpoint is not None:
            r = requests.get(endpoint, headers=auth_headers)
            if r.status_code != 200:
                r.raise_for_status()
            response = json.loads(r.text)
            for r in response['results']:
                filtered_datasets[r['name']] = r['id']
            endpoint = response['next']
    return filtered_datasets


@fetch_auth
def create_generated_dataset(name, sim_name, config, url, auth_headers):
    """ create dataset

    Create generated dataset on ZumoLabs backend which will launch
    a generation job with specified params.

    Args:
        name (str): name of dataset
        sim_name (str): name of sim the dataset is built from
        config (dict): configration of sim for this dataset
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f'{url}/api/v1/sims/'
    r = requests.get(
        endpoint, 
        params={'name': sim_name},
        headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    response = json.loads(r.text)
    if response['count'] != 1:
        raise NameError(f"found {response['count']} sims for name {sim_name}")
    sim = response['results'][0]
    endpoint = f'{url}/api/v1/generated-data-sets/'
    r = requests.post(
        endpoint, 
        data={
            'sim': sim['id'],
            'config': json.dumps(config),
            'name': name
        }, 
        headers=auth_headers)
    if r.status_code != 201:
        r.raise_for_status()


@fetch_auth
def create_uploaded_dataset(name, path, url, auth_headers):
    """ upload dataset

    Upload dataset to S3 through ZumoLabs backend and 
    create object.

    Args:
        name (str): name of dataset
        path (str): path to dataset
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f'{url}/api/v1/uploaded-data-sets/'
    r = requests.post(
        endpoint, 
        data={'name': name},
        files={'file': open(path, 'rb')},
        headers=auth_headers)
    if r.status_code != 201:
        r.raise_for_status()


@fetch_auth
def download_dataset(name, path, dataset_type, url, auth_headers):
    """ download dataset

    Download dataset object from S3 through ZumoLabs backend.

    Args:
        name (str): name of dataset to download
        path (str): output directory
        dataset_type (str): type of dataset to download
        url (str): backend endpoint
        auth_headers: authentication for backend

    Returns:
        str: output file path
    """
    endpoint = f'{url}/api/v1/{dataset_type}-data-sets/'
    r = requests.get(endpoint, params={'name': name}, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    response = json.loads(r.text)
    if response['count'] != 1:
        raise NameError(f"found {response['count']} datasets for name {name}")
    dataset = response['results'][0]
    endpoint = f"{url}/api/v1/{dataset['dataset_type']}-data-sets/{dataset['id']}/download"
    r = requests.get(endpoint, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    response = json.loads(r.text)
    name_slug = f"{dataset['name'].replace(' ', '_')}-{dataset['id'][:8]}.zip"
    output_path = to_pathlib_path(path) / name_slug
    download_url(response['redirect_link'], output_path)
    return output_path


def fetch_datasets():
    """ fetch datasets

    Fetch dataset objects from ZumoLabs backend. Fetches all three 
    datasets types uploaded, generated, and job.

    Returns:
        list: paginated sorted datasets for all types
    """
    datasets = []
    for dataset_type in DATASET_TYPES:
        short_dataset_type = dataset_type.split('-')[0]
        datasets += [(lambda d: d.update({'type': short_dataset_type}) or d)(d) 
            for d in _fetch_type_datasets(dataset_type)]
    return sorted(datasets, key = lambda i: i['created_at'], reverse=True)


@fetch_auth
def _fetch_type_datasets(dataset_type, url, auth_headers):
    """ fetch type of datasets

    Args:
        dataset_type (str): type of dataset to fetch
        url (str): backend endpoint
        auth_headers: authentication for backend

    Returns:
        list: paginated datasets for given type
    """
    endpoint = f'{url}/api/v1/{dataset_type}/'
    r = requests.get(endpoint, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    return json.loads(r.text)['results']



"""
    Machine Learning Experiment Hooks.
"""
from typing import Dict, Union
from pathlib import Path
import json
import requests
import logging

ENDPOINT = 'https://ragnarok.zumok8s.org/api/v1/experiment/'
experiment = None
logger = None

def init(
    name: str,
    sim: str = None,
    dataset: str = None,
    config: Dict = None,
    api_key: str = None,
) -> None:
    """ Initialize a experiment run.

    Args:
        name (str): identifier for experiment
        sim (str. optional): identifier for simulation associated with experiment
        dataset (str, optional): identifier for dataset associated with experiment
        config (dict, optional): configuration details about experiment
        api_key (str, required): api_key to auth with backend
    Returns:
        experiment object
    """
    if api_key is None:
        raise PermissionError('please input zpy api_key')
    global logger
    logger = logging.getLogger(__name__)
    exp = Experiment(name=name, sim=sim, dataset=dataset, config=config, api_key=api_key)
    global experiment
    experiment = exp
    exp._create()


def log(
    metrics: str = None,
    file_path: str = None,
) -> None:
    """ Log an update to experiment.

    Args:
        metrics (str, optional): free form dictionary of data to log
        file_path (str, optional): file path to upload

    Raises:
        FileNotFoundError: if file_path doesnt exist
    """
    global experiment
    exp = experiment
    if file_path:
        file_path = Path(file_path).resolve()
    exp._update(file_path=file_path, metrics=metrics)


class Experiment:
    """ experiment class for uploading """

    def __init__(
            self,
            name: str = None,
            sim: str = None,
            dataset: str = None,
            config: Dict = None,
            api_key: str = None
        ) -> None:
        """ Experiment Class for ragnarok upload.

        Args:
            name (str): identifier for experiment
            sim (str. optional): identifier for simulation associated with experiment
            dataset (str, optional): identifier for dataset associated with experiment
            config (dict, optional): configuration details about experiment
            api_key (str, required): api_key to auth with backend
        """
        self.name = name
        self.sim = sim
        self.dataset = dataset
        self.config = config
        self.auth_headers = {'Authorization': 'token {}'.format(api_key)}
        self.id = None

    def _post(self, data=None):
        """ post to endpoint """
        r = requests.post(ENDPOINT, data=data, headers=self.auth_headers)
        if r.status_code != 201:
            logger.debug(f'{r.text}')
            r.raise_for_status()
        self.id = json.loads(r.text)['id']
        logger.debug(f'{r.status_code}: {r.text}')

    def _put(self, data=None, files=None):
        """ put to endpoint """
        r = requests.put(f'{ENDPOINT}{self.id}/', data=data, files=files, headers=self.auth_headers)
        if r.status_code != 200:
            logger.debug(f'{r.text}')
            r.raise_for_status()
        logger.debug(f'{r.status_code}: {r.text}')

    def _create(self):
        """ request to create experiment """
        data = {'name': self.name}
        if self.sim:
            data['sim_name'] = self.sim
        if self.dataset:
            data['data_set_name'] = self.dataset
        if self.config:
            data['config'] = json.dumps(self.config)
        self._post(data=data)

    def _update(
            self,
            file_path: Union[Path, str] = None,
            metrics: Dict = None
        ) -> None:
        """ request to update experiment """
        data = {'name': self.name}
        if metrics:
            data['metrics'] = json.dumps(metrics)
        files = None
        if file_path:
            files = {'file': open(file_path, 'rb')}
            data['file_name'] = Path(file_path).name
        self._put(data=data, files=files)
